/*  RAxML-VI-HPC (version 2.2) a program for sequential and parallel estimation of phylogenetic trees 
 *  Copyright August 2006 by Alexandros Stamatakis
 *
 *  Partially derived from
 *  fastDNAml, a program for estimation of phylogenetic trees from sequences by Gary J. Olsen
 *  
 *  and 
 *
 *  Programs of the PHYLIP package by Joe Felsenstein.
 *
 *  This program is free software; you may redistribute it and/or modify its
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 * 
 *
 *  For any other enquiries send an Email to Alexandros Stamatakis
 *  Alexandros.Stamatakis@epfl.ch
 *
 *  When publishing work that is based on the results from RAxML-VI-HPC please cite:
 *
 *  Alexandros Stamatakis:"RAxML-VI-HPC: maximum likelihood-based phylogenetic analyses with thousands of taxa and mixed models". 
 *  Bioinformatics 2006; doi: 10.1093/bioinformatics/btl446
 */

#ifndef WIN32
#include <sys/times.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h> 
#endif

#include <math.h>
#include <time.h> 
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

#include "axml.h"

extern const int mask32[32];
extern int Thorough;

static void findNextRec(nodeptr p, tree *tr, int *count, nodeptr *nodes, int depth)
{
  int 
    model,
    notFound;

  if(isTip(p->number, tr->mxtips))
    {
      for(model = 0; model < tr->NumberOfModels; model++)
	{
	  if(!nodes[model] && (p->isPresent[model / MASK_LENGTH] & mask32[model % MASK_LENGTH]))
	    {
	      assert(p->backs[model]);
	      nodes[model] = p;
	      count[model] = depth;
	    }
	}
    }
  else
    {
      assert(p == p->next->next->next);

      for(model = 0, notFound = 0; model < tr->NumberOfModels; model++)
	{
	  if(!nodes[model])
	    {
	      if(p->backs[model])
		{
		  nodes[model] = p;
		  count[model] = depth;	       
		}
	      else
		notFound++;
	    }
	  
	}

      if(notFound == 0)
	return;

      findNextRec(p->next->back, tr, count, nodes, depth + 1);
      findNextRec(p->next->next->back, tr, count, nodes, depth + 1);
    }

}

void findNext(nodeptr p, tree *tr, nodeptr *result)
{
  int 
    depth,
    model,
    leftDepth[NUM_BRANCHES],
    rightDepth[NUM_BRANCHES];
  
  nodeptr
    leftNodes[NUM_BRANCHES],
    rightNodes[NUM_BRANCHES];

  for(model = 0; model < tr->NumberOfModels; model++)
    {
      leftDepth[model]  = -1;
      rightDepth[model] = -1;

      result[model]     = (nodeptr)NULL;      
      leftNodes[model]  = (nodeptr)NULL;
      rightNodes[model] = (nodeptr)NULL;
    }

  depth = 0;
  findNextRec(p, tr, leftDepth, leftNodes, 0);

  depth = 0; 
  findNextRec(p->back, tr, rightDepth, rightNodes, 0);


  for(model = 0; model < tr->NumberOfModels; model++)
    {
      assert(rightDepth[model] >= 0 || leftDepth[model] >= 0);
      assert(rightNodes[model] || leftNodes[model]);

      if(rightDepth[model] >= 0 && leftDepth[model] >= 0)
	{
	  if(rightDepth[model] < leftDepth[model])
	    result[model] = rightNodes[model];
	  else 
	    result[model] = leftNodes[model];
	}
      else
	{
	  if(rightDepth[model] >= 0)
	    result[model] = rightNodes[model];
	  else
	    result[model] = leftNodes[model];
	}
	 
      assert(result[model]->backs[model] && result[model]->next->backs[model] && result[model]->next->next->backs[model]);
      /* printf("Partition %d node %d\n", model, result[model]->number); */
    
    }
}


int containsModel(nodeptr p, int model, tree *tr)
{
  if(isTip(p->number, tr->mxtips))
    {
      if(p->isPresent[model / MASK_LENGTH] & mask32[model % MASK_LENGTH])
	return 1;
      else
	return 0;
    }
  else
    {
      assert(p = p->next->next->next);
      return (containsModel(p->next->back, model, tr) || containsModel(p->next->next->back, model, tr));
    }
}

static void reduceTreeModelREC(nodeptr p, nodeptr reference, int model, tree *tr, int *branchCounter)
{

  if(isTip(p->number, tr->mxtips))
    {
      assert(p->isPresent[model / MASK_LENGTH] & mask32[model % MASK_LENGTH]);

      p->backs[model]         = reference;
      reference->backs[model] = p;
      p->z[model] = reference->z[model] = defaultz;

      *branchCounter = *branchCounter + 1;

    }
  else
    {
      nodeptr q = p->next;
      nodeptr r = p->next->next;
      
      int left  = containsModel(q->back, model, tr);      
      int right = containsModel(r->back, model, tr);

      assert(p = p->next->next->next);

      if(left && right)
	{
	  p->backs[model]         = reference;
	  reference->backs[model] = p;
	  p->z[model] = reference->z[model] = defaultz;

	  *branchCounter = *branchCounter + 1;

	  reduceTreeModelREC(q->back, q,  model, tr, branchCounter);   	 	
	  reduceTreeModelREC(r->back, r,  model, tr, branchCounter);
	} 
      else
	{
	  if(left || right)
	    {
	      if(left)		
		reduceTreeModelREC(q->back, reference,  model, tr, branchCounter);
	      else	
		reduceTreeModelREC(r->back, reference,  model, tr, branchCounter);		 
	    }
	  else	    
	    assert(0);	      	    
	}     
    }
}





/*static void initravPresence(nodeptr p, int ntips)
{
  if(!(isTip(p->number, ntips)))
    {
      int i;

      nodeptr 
	l = p->next->back,
	r = p->next->next->back;

      assert(p->next->next->next == p);

      initravPresence(l, ntips);
      initravPresence(r, ntips);
	
      for(i = 0; i < VECTOR_LENGTH; i++)
	p->isPresent[i] = l->isPresent[i] | r->isPresent[i];
    }

    }*/


static void recAssign(tree *tr, int model, size_t entries, size_t width, nodeptr p)
{
  if(isTip(p->number, tr->mxtips))
    {
      assert(p->isPresent[model / MASK_LENGTH] & mask32[model % MASK_LENGTH]);
    }
  else
    {
      nodeptr q;

      /* assign xVector, xVector_FLOAT and conditional scaling array */
      
      assert(p->backs[model]);

      q = p->next;

      if(tr->useFloat)
	tr->partitionData[model].xVector_FLOAT[p->number - tr->mxtips -1] = (float*)malloc(sizeof(float) * entries);
      else
	tr->partitionData[model].xVector[p->number - tr->mxtips -1] = (double*)malloc(sizeof(double) * entries);

      if(!tr->useFastScaling)
	tr->partitionData[model].expVector[p->number - tr->mxtips - 1] = (int *)malloc(sizeof(int) * width);

      while(q != p)
	{
	  assert(q->backs[model]);
	  recAssign(tr, model, entries, width, q->backs[model]);
	  q = q->next;
	}

    }
}


static void allocNodexMesh (tree *tr)
{
  size_t    
    model;    

  for(model = 0; model < (size_t)tr->NumberOfModels; model++)
    {
      size_t 
	width = tr->partitionData[model].upper - tr->partitionData[model].lower,     
	entries = (size_t)(tr->discreteRateCategories) * (size_t)(tr->partitionData[model].states) * width;      

      recAssign(tr, model, entries, width, tr->startVector[model]->backs[model]);
    } 
}


void setupPointerMesh(tree *tr)
{
  if(tr->multiGene)
    {
      int model;
      
      assert(isTip(tr->start->number, tr->mxtips));
      printf("Global start at tip %d\n", tr->start->number);
      
      /*initravPresence(tr->start->back, tr->mxtips);*/
      
      for(model = 0; model < tr->NumberOfModels; model++)
	{ 
	  int branchCounter = 0;
	  assert(isTip(tr->startVector[model]->number, tr->mxtips));
	  
	  reduceTreeModelREC(tr->startVector[model]->back, tr->startVector[model], model, tr, &branchCounter);
	  
	  printf("Partition %d has %d branches\n", model, 2 * tr->mxtipsVector[model] - 3);
	  assert(branchCounter == 2 * tr->mxtipsVector[model] - 3);
	  
	}
      
      allocNodexMesh(tr);
    }
}


nodeptr getPartition(int model, tree *tr, nodeptr p)
{
  if(isTip(p->number, tr->mxtips))
    {
      if(p->isPresent[model / MASK_LENGTH] & mask32[model % MASK_LENGTH])
	{
	  assert(p->backs[model]);	  	  	  
	  return p;
	} 
      else
	return ((nodeptr)NULL);
    }
  else
    {
      if(p->backs[model])
	{
	  assert(p->next->backs[model] && p->next->next->backs[model]);	
	  return p;
	}
      else
	{	 
	  nodeptr 
	    r1 = (nodeptr)NULL, 
	    r2 = (nodeptr)NULL;

	  r1 = getPartition(model, tr, p->next->back); 	  	 
	  r2 = getPartition(model, tr, p->next->next->back);
	  
	  assert(!(r1 && r2));

	  if(r1)
	    {
	      assert(r1->backs[model] != p->next);
	      return r1;
	    }
	  else
	    {
	      if(r2)
		{
		  assert(r2->backs[model] != p->next->next);
		  return r2;
		}
	      else
		return ((nodeptr)NULL);
	    }
	}

    }

}


void determineSubtreeAndNeighbors(tree *tr, nodeptr p)
{
  /* q---p---r  */
  /*     |      */
  /*   p->Back  */

  int model; 

  for(model = 0; model < tr->NumberOfModels; model++)
    {
      /* init to NULL, find remove node for all partitions */

      tr->removeNodes[model] = (nodeptr)NULL;

      if(p->backs[model])
	{
	  /* this is the easy case all departing branches of node 
	     p contain taxa of for partition model 
	  */

	  assert(p->next->backs[model] && p->next->next->backs[model]);
	  tr->removeNodes[model] = p;	 

	  {
	    /* some extra assertions to feel better */

	    int 
	      subtree = containsModel(p->back, model, tr),
	      left = containsModel(p->next->back, model, tr),
	      right = containsModel(p->next->next->back, model, tr);

	    assert(subtree && left && right);
	  }
	 
	}
      else
	{ 
	  int 
	    subtree,
	    left,
	    right;

	  /* check if the subtree to be pruned actually contains 
	     taxa of partition model */

	  subtree = containsModel(p->back, model, tr);
	 

	  /* if not we don't care at all about this partition */

	  if(subtree)
	    {
	      left    = containsModel(p->next->back, model, tr);
	      right   = containsModel(p->next->next->back, model, tr);

	      /*
		either none of the trees to the left and right of the 
		 pruning position contains a taxon for partition model in which case 
		 we are done.
		 otherwise the left XOR the right subtree must contain taxa for our partition 
	      */

	      assert((!left && !right) || (left && !right) || (!left && right));

	      if(left || right)
		{		 		  
		  nodeptr q = getPartition(model, tr, p->back);
		  assert(q);
		  assert(q->backs[model]);
		  assert(q->backs[model]->backs[model]);
		  assert(q == q->backs[model]->backs[model]);

		  /* if the removal node is a tip there is no work to do ! */

		  if(!isTip(q->backs[model]->number, tr->mxtips))		    		    		 		 
		    tr->removeNodes[model] = q->backs[model];
		}
	    }	 	  
	}	  
    }
}


static void hookup_MULTI(nodeptr p, nodeptr q)
{
  p->back = q;
  q->back = p;
}

static void hookupMESH(nodeptr p,  nodeptr q, double z, int model)
{
  p->backs[model] = q;
  q->backs[model] = p;
  q->z[model] = p->z[model] = z;
}

static double insertBIG_MESH (tree *tr, nodeptr q, nodeptr s, nodeptr p, int model)
{
  /* insert q ----p-----s */
  /*              |       */
  /*              r       */
  
  double result = 0.0;
  int i;                  
  double  z; 
  
  z = sqrt(q->z[model]);      
  
  if(z < zmin) 
    z = zmin;
  if(z > zmax)
    z = zmax;
  
  hookupMESH(p->next,       q, z, model);
  hookupMESH(p->next->next, s, z, model);	                         
  
  for(i = 0; i < tr->NumberOfModels; i++)
    if(i != model)
      tr->executeModel[i] = FALSE;
    else
      tr->executeModel[i] = TRUE;
  
  
  newviewGeneric(tr, p);   
  
  result = evaluateGenericMulti(tr, p->next->next, model);

  

  printf("%d %f\n", model, result);
  
  for(i = 0; i < tr->NumberOfModels; i++)
    tr->executeModel[i] = TRUE;
  
  return  result;
}


static void  removeNodeBIG_MULTI(tree *tr, nodeptr p, int numBranches)
{  
  double   zqr[NUM_BRANCHES], result[NUM_BRANCHES];
  nodeptr qGlobal, rGlobal;
  int model;
        
  for(model = 0; model < tr->NumberOfModels; model++)
    {
      tr->leftNodes[model]   = (nodeptr)NULL;
      tr->rightNodes[model]  = (nodeptr)NULL;
      tr->removeNodes[model] = (nodeptr)NULL;
    }
      
  determineSubtreeAndNeighbors(tr, p);
  
  for(model = 0; model < tr->NumberOfModels; model++)
    {
      if(tr->removeNodes[model])
	{
	  nodeptr  q, r, reference;

	  reference = tr->removeNodes[model];

	  assert(!isTip(reference->number, tr->mxtips));
	  assert(reference->backs[model] && reference->next->backs[model] && reference->next->next->backs[model]);

	  q = reference->next->backs[model];
	  r = reference->next->next->backs[model];

	  assert(q->backs[model] && r->backs[model]);
	  assert(q->next->backs[model] && r->next->backs[model]);
	  assert(q->next->next->backs[model] && r->next->next->backs[model]);

	  assert(q->z[model] == reference->next->z[model]);
	  assert(r->z[model] == reference->next->next->z[model]);
	  zqr[model] = q->z[model] * r->z[model];
	  printf("Model %d %f\n", model, zqr[model]);

	  tr->leftNodes[model] = q;
	  tr->rightNodes[model] = r;
	  	  
	  assert(tr->rightNodes[model]->backs[model] == reference->next->next && 
		 tr->leftNodes[model]->backs[model] == reference->next);

	}     
    }

  printf("\n\n"); 

  for(model = 0; model < tr->NumberOfModels; model++)
    {
      int i;

      tr->storedPerPartitionLH[model] = tr->perPartitionLH[model];
      printf("P %d: %f\n", model, tr->storedPerPartitionLH[model]);

      for(i = 0; i < tr->NumberOfModels; i++)
	tr->partitionConverged[i] = TRUE;

      if(tr->removeNodes[model])
	{
	  nodeptr  q, r, reference;

	  printf("%d %f -> ", model, zqr[model]);
	  reference = tr->removeNodes[model];
	  q = reference->next->backs[model];
	  r = reference->next->next->backs[model];
	  tr->partitionConverged[model] = FALSE;
	  makenewzGeneric(tr, q, r, zqr, iterations, result, TRUE);   
	  printf("%f \n", result[model]);

	  assert(!isTip(reference->number, tr->mxtips));

	  q->backs[model] = r;
	  r->backs[model] = q;

	  tr->zLeft[model]  = q->z[model];
	  tr->zRight[model] = r->z[model];
	  tr->zDown[model]  = reference->z[model];

	  
	  r->z[model] = q->z[model] == result[model];

	  assert(r->backs[model] == q);
	  assert(q->backs[model] == r);

	  reference->next->backs[model] = reference->next->next->backs[model] = (nodeptr)NULL;
	}
    }  
    
  qGlobal = p->next->back;
  rGlobal = p->next->next->back;

  hookup_MULTI(qGlobal, rGlobal);
    
  p->next->next->back = p->next->back = (nodeptr)NULL; 
}

int skip = 0;
int execute = 0;
int insertions = 0;

static void testInsertBIG_MULTI (tree *tr, nodeptr p, nodeptr q)
{
  double  qz[NUM_BRANCHES], pz[NUM_BRANCHES];
  nodeptr  r;
  boolean doIt = TRUE;
  double startLH = tr->endLH;
  int i;
  double likelihood = 0.0;
  
  /* insert p between r and q :-) */

  r = q->back;   
  
  printf("Insert Node: %d between %d and %d\n", p->number, r->number, q->number);

  if(tr->grouped)
    {
      int rNumber, qNumber, pNumber;
      
      doIt = FALSE;
      
      rNumber = tr->constraintVector[r->number];
      qNumber = tr->constraintVector[q->number];
      pNumber = tr->constraintVector[p->number];
      
      if(pNumber == -9)
	pNumber = checker(tr, p->back);
      if(pNumber == -9)
	doIt = TRUE;
      else
	{
	  if(qNumber == -9)
	    qNumber = checker(tr, q);
	  
	  if(rNumber == -9)
	    rNumber = checker(tr, r);
	  
	  if(pNumber == rNumber || pNumber == qNumber)
	    doIt = TRUE;       
	}
    }
  
  if(doIt)
    {     
      int model;
      boolean executeIt[NUM_BRANCHES];     
      
      for(model = 0; model < tr->NumberOfModels; model++)
	executeIt[model] = FALSE;

      for(model = 0; model < tr->NumberOfModels; model++)
	{
	  if(tr->removeNodes[model])
	    {
	      int left, right;
	      nodeptr 
		qModel = (nodeptr)NULL,
		rModel = (nodeptr)NULL;
	      
	      printf("Model %d\n", model);

	      assert(!isTip(tr->removeNodes[model]->number, tr->mxtips));

	      printf("REMOVE: %d \n", tr->removeNodes[model]->number);


	      /* 
		 What happens when either q or r is equal to  tr->removeNodes[model]
		 it should be skipped anyway by our mechanism since the nieghboring nodes will be 
		 the ones of the reomval position anyway 
	      */

	      if(q->backs[model])
		{
		  printf("Q: %d \n", q->number);

		  qModel = q;
		  if(!isTip(qModel->number, tr->mxtips) && (tr->removeNodes[model]->number != q->number))
		    assert(q->next->backs[model] && q->next->next->backs[model]);
		  
		  if(isTip(qModel->number, tr->mxtips))
		    assert(qModel->isPresent[model / MASK_LENGTH] & mask32[model % MASK_LENGTH]);
		    
		  left = 1;
		}
	      else
		left = containsModel(q, model, tr);
	      
	      if(r->backs[model])
		{
		  printf("R: %d \n", r->number);

		  rModel = r;
		  if(!isTip(rModel->number, tr->mxtips) && (tr->removeNodes[model]->number != r->number) )
		    assert(r->next->backs[model] && r->next->next->backs[model]);
		  
		  if(isTip(rModel->number, tr->mxtips))
		    assert(rModel->isPresent[model / MASK_LENGTH] & mask32[model % MASK_LENGTH]);

		  right = 1;
		}
	      else
		right = containsModel(r, model, tr);

	      if(left && right)
		{
		  if(!qModel)		    		      
		    qModel = getPartition(model, tr, q);
		    
		  
		  if(!rModel)
		    rModel = getPartition(model, tr, r);

		  assert(rModel && qModel);

		  assert(rModel->backs[model] == qModel);
		  assert(qModel->backs[model] == rModel);

		  if((rModel == tr->leftNodes[model] && qModel == tr->rightNodes[model]) ||
		     (rModel == tr->rightNodes[model] && qModel == tr->leftNodes[model]))		     
		    {
		      likelihood += tr->storedPerPartitionLH[model];
		      printf("Skip %d %d %f\n", model, skip, likelihood);
		      skip++;
		    }
		  else
		    {
		      /* re-compute likelihood here for specific model */
		      double z = qModel->z[model];
		      assert(qModel->z[model] == rModel->z[model]);
		      likelihood += insertBIG_MESH(tr, qModel, rModel, tr->removeNodes[model], model);
		      printf("Execute %d %d %f\n", model, execute, likelihood);
		      hookupMESH(qModel, rModel, z, model);
		      tr->removeNodes[model]->next->next->backs[model] = tr->removeNodes[model]->next->backs[model] = (nodeptr) NULL;
		      execute++;
		    }
		}
	      else
		{ 
		  likelihood += tr->storedPerPartitionLH[model];
		  printf("Skip %d %d %f\n", model, skip, likelihood);
		  skip++;
		}

	    }
	  else
	    {
	      likelihood += tr->storedPerPartitionLH[model];
	      printf("Skip %d %d %f\n", model, skip, likelihood);
	      skip++;
	    }
	}

      tr->likelihood = likelihood;

     
      printf("Insertion %d: %f\n", insertions, tr->likelihood);

      insertions++;
      if(insertions == 10)
	exit(1);     
      
      hookup_MULTI(q, r);
      
      p->next->next->back = p->next->back = (nodeptr) NULL;
      
     
      return;
    }
  else
    return;  
}


static void addTraverseBIG_MULTI(tree *tr, nodeptr p, nodeptr q, int mintrav, int maxtrav)
{  
  
  if (--mintrav <= 0) 
    {
      testInsertBIG_MULTI(tr, p, q);      
    }
  
  if ((!isTip(q->number, tr->rdta->numsp)) && (--maxtrav > 0)) 
    {    
      addTraverseBIG_MULTI(tr, p, q->next->back, mintrav, maxtrav);
      addTraverseBIG_MULTI(tr, p, q->next->next->back, mintrav, maxtrav);    
    }
} 

int rearrangeBIG_MULTI(tree *tr, nodeptr pGlobal, int mintrav, int maxtrav)   
{  
  double   p1z[NUM_BRANCHES], p2z[NUM_BRANCHES], q1z[NUM_BRANCHES], q2z[NUM_BRANCHES];
  nodeptr  p1, p2, qGlobal, q1, q2;
  int      mintrav2, i;  
  boolean doP = TRUE, doQ = TRUE;
  
  if (maxtrav < 1 || mintrav > maxtrav)  return 0;
  qGlobal = pGlobal->back;
  
  if(tr->constrained)
    {
      if(! tipHomogeneityChecker(tr, pGlobal->back, 0))
	doP = FALSE;
      
      if(! tipHomogeneityChecker(tr, qGlobal->back, 0))
	doQ = FALSE;
      
      if(doQ == FALSE && doP == FALSE)
	return 0;
    }
  
  if (!isTip(pGlobal->number, tr->rdta->numsp) && doP) 
    {     
      p1 = pGlobal->next->back;
      p2 = pGlobal->next->next->back;
      
     
      if(!isTip(p1->number, tr->rdta->numsp) || !isTip(p2->number, tr->rdta->numsp))
	{
	  printf("Front %d\n", pGlobal->number);
	 	  
	  removeNodeBIG_MULTI(tr, pGlobal,  tr->numBranches);
	  	  
	  
	  if(!isTip(p1->number, tr->rdta->numsp)) 
	    {
	      addTraverseBIG_MULTI(tr, pGlobal, p1->next->back,
				  mintrav, maxtrav);         
	      addTraverseBIG_MULTI(tr, pGlobal, p1->next->next->back,
				   mintrav, maxtrav);          
	    }
	    
	  if(!isTip(p2->number, tr->rdta->numsp)) 
	    {
	      addTraverseBIG_MULTI(tr, pGlobal, p2->next->back,
				   mintrav, maxtrav);
	      addTraverseBIG_MULTI(tr, pGlobal, p2->next->next->back,
				   mintrav, maxtrav);          
	    }
	  
	  	  
	  hookup_MULTI(pGlobal->next,       p1); 
	  hookup_MULTI(pGlobal->next->next, p2);	   	    	    
	  
	  {	  
	    int model;
	    for(model = 0; model < tr->NumberOfModels; model++)
	      {
		if(tr->removeNodes[model])
		  {
		    nodeptr 
		      reference = tr->removeNodes[model],
		      q = tr->leftNodes[model],
		      r = tr->rightNodes[model],
		      s = reference->backs[model];

		    assert(reference && q && r && s);
		    assert(reference->backs[model] == s);
		    assert(s->backs[model] == reference);

		    if(!isTip(q->number, tr->mxtips))
		      assert(q->backs[model] && q->next->backs[model] && q->next->next->backs[model]);
		    if(!isTip(r->number, tr->mxtips))
		      assert(r->backs[model] && r->next->backs[model] && r->next->next->backs[model]);
		    if(!isTip(s->number, tr->mxtips))
		      assert(s->backs[model] && s->next->backs[model] && s->next->next->backs[model]);
		    if(!isTip(reference->number, tr->mxtips))
		      assert(reference->backs[model]);

		    reference->z[model] = s->z[model]             = tr->zDown[model];
		    q->z[model] = reference->next->z[model]       = tr->zLeft[model];
		    r->z[model] = reference->next->next->z[model] = tr->zRight[model];
		    
		    q->backs[model] = reference->next;
		    reference->next->backs[model] = q;

		    r->backs[model] = reference->next->next;
		    reference->next->next->backs[model] = r;

		    if(!isTip(reference->number, tr->mxtips))
		      assert(reference->backs[model] && reference->next->backs[model] && reference->next->next->backs[model]);


		  }		
	      }
	    
	  }


	  /*
	    initravDIST(tr, p, 1);	   	    
	  */
	}
    }
  
  if (!isTip(qGlobal->number, tr->rdta->numsp) && maxtrav > 0 && doQ) 
    {
      q1 = qGlobal->next->back;
      q2 = qGlobal->next->next->back;
      
          
      if (
	  (
	   ! isTip(q1->number, tr->rdta->numsp) && 
	   (! isTip(q1->next->back->number, tr->rdta->numsp) || ! isTip(q1->next->next->back->number, tr->rdta->numsp))
	   )
	  ||
	  (
	   ! isTip(q2->number, tr->rdta->numsp) && 
	   (! isTip(q2->next->back->number, tr->rdta->numsp) || ! isTip(q2->next->next->back->number, tr->rdta->numsp))
	   )
	  )
	{	 
	  printf("Back %d\n", qGlobal->number);
   	 	  
	  removeNodeBIG_MULTI(tr, qGlobal, tr->numBranches);
	  
	  
	  mintrav2 = mintrav > 2 ? mintrav : 2;
	  
	  if ( !isTip(q1->number, tr->rdta->numsp)) 
	    {
	      addTraverseBIG_MULTI(tr, qGlobal, q1->next->back,
				   mintrav2 , maxtrav);
	      addTraverseBIG_MULTI(tr, qGlobal, q1->next->next->back,
				   mintrav2 , maxtrav);         
	    }
	    
	  if ( ! isTip(q2->number, tr->rdta->numsp)) 
	    {
	      addTraverseBIG_MULTI(tr, qGlobal, q2->next->back,
				   mintrav2 , maxtrav);
	      addTraverseBIG_MULTI(tr, qGlobal, q2->next->next->back,
				   mintrav2 , maxtrav);          
	    }	
	  
	   {
	    /* repair branches and mesh */

	    int model;
	    for(model = 0; model < tr->NumberOfModels; model++)
	      {
		if(tr->removeNodes[model])
		  {
		    nodeptr 
		      reference = tr->removeNodes[model],
		      q = tr->leftNodes[model],
		      r = tr->rightNodes[model],
		      s = reference->backs[model];

		    assert(reference && q && r && s);
		    assert(reference->backs[model] == s);
		    assert(s->backs[model] == reference);

		    if(!isTip(q->number, tr->mxtips))
		      assert(q->backs[model] && q->next->backs[model] && q->next->next->backs[model]);
		    if(!isTip(r->number, tr->mxtips))
		      assert(r->backs[model] && r->next->backs[model] && r->next->next->backs[model]);
		    if(!isTip(s->number, tr->mxtips))
		      assert(s->backs[model] && s->next->backs[model] && s->next->next->backs[model]);
		    if(!isTip(reference->number, tr->mxtips))
		      assert(reference->backs[model]);

		    reference->z[model] = s->z[model]             = tr->zDown[model];
		    q->z[model] = reference->next->z[model]       = tr->zLeft[model];
		    r->z[model] = reference->next->next->z[model] = tr->zRight[model];
		    
		    q->backs[model] = reference->next;
		    reference->next->backs[model] = q;

		    r->backs[model] = reference->next->next;
		    reference->next->next->backs[model] = r;

		     if(!isTip(reference->number, tr->mxtips))
		      assert(reference->backs[model] && reference->next->backs[model] && reference->next->next->backs[model]);

		  }
		
	      }
	    
	  }
	    
	  hookup_MULTI(qGlobal->next,       q1); 
	  hookup_MULTI(qGlobal->next->next, q2);
	    
	    /*
	    initravDIST(tr, q, 1); 	  
	  */
	}
    }
  
 
  return  1;
} 

